// code by jph

#define  ANIPATCH_PUSH   0.05


class aniFlow : public aniMated {
public:
  aniFlow(char* pfn,int level);
  ~aniFlow(void);
  void scene(void);
  void draw(void);
  
  char* toString(void) { return "aniFlow"; };
  linked<aniPatch*> list;
  tensor<double>* fld;
  blue<double>* blu;
  tensor<char> str;
  int ord;
  int lev;
};

aniFlow::aniFlow(char* pfn,int level) {
  fld=new tensor<double>[2];
  fld[0]=tensor<double>("get/flow/field_0.arr");
  fld[1]=tensor<double>("get/flow/field_1.arr");
  blu=new blue<double>(tensor<int>::vector(2).as(45,60),0.8);
  tensor<double> pnt(1,2);
  tensor<double> res(1,2);
  for (int c0=0;c0<blu->grid.size[0];++c0)
  for (int c1=0;c1<blu->grid.size[1];++c1) {
    pnt(0)=c0/(double)blu->grid.size[0];
    pnt(1)=c1/(double)blu->grid.size[1];
    res(0)=fld[0].interp(pnt);
    res(1)=fld[1].interp(pnt);
    blu->grid(c0,c1)=-(res.norm()<TENSOR_NO_DIVISION); // TODO simplify       
  }
  //tensor<double> ext=blu->extension();
  //printf("FIELD EXTENSION %s\n",ext.toString());
    
  str=tensor<char>::string(pfn);
  ord=0;
  lev=level;
  
  //pnt.as( 22.0, 5.0);
  //aniPatch* ptc;
  //if (blu->insert(pnt)) {
  //  printf("yeay!\n");
  //  ptc=new aniPatch(str,blu->rad,pnt,ord++,lev);
  //  ptc->psh=+ANIPATCH_PUSH;
  //  list.add(ptc);
  //}
  //printf("%s\n",blu->toInfo());
  //blu->complete(20);
  //tensor<double> arr=blu->point.toTensor(2);
  //aniPatch* ptc;
  //for (int c0=0;c0<arr.size[1];++c0) {
  //  ptc=new aniPatch(str,blu->rad,arr[c0],ord++,lev);
  //  ptc->alp=1;
  //  list.add(ptc);
  //}
  //blu->empty();
}

aniFlow::~aniFlow(void) {
  delete fld;
}

void aniFlow::scene(void) {
  int c0,c1;
  aniPatch* ptc;
  meshgl* msh;
  tensor<double> pnt(1,2);
  tensor<double> ext=blu->extension();
  ext(0)=1/ext(0);
  ext(1)=1/ext(1);
  int v3f;
  if (!keyboard('e')||opengl_recording) {
    time tic=time::now();      
    tensor<aniPatch*> crt;
    tensor<double> nrm;
    crt=list.toTensor();
    list.empty();
    nrm=tensor<double>(1,crt.numel);     
    for (c0=0;c0<crt.numel;++c0)
      nrm(c0)=crt(c0)->norm();
    tensor<int> ind;
    nrm.sort(TENSOR_SORT_ABC,ind); // sort patches according to deformation: min to max
    if (crt.numel)
      v3f=crt(0)->msh->offsetV3F();
    for (c0=0;c0<crt.numel;++c0) { // for all patches
      ptc=crt(ind(c0));
      msh=ptc->msh;
      // move all mesh points with flow      
      //aniMated::dt=0.22;
      for (c1=v3f;c1<msh->arr.numel;c1+=msh->arr.size[0]) {
        pnt.as( (double)msh->arr.data[c1+0],
                (double)msh->arr.data[c1+1]);
        pnt%=ext;
        msh->arr.data[c1+0]+=fld[0].interp(pnt)*aniMated::dt; // move points of patch mesh
        msh->arr.data[c1+1]+=fld[1].interp(pnt)*aniMated::dt; // move points of patch mesh
      }
      if (0<=ptc->psh) // if increment is non-negative
        if (nrm(ind(c0))<2) { // 2
          ptc->psh=ANIPATCH_PUSH;
          pnt.as( (double)msh->arr.data[v3f+0],
                  (double)msh->arr.data[v3f+1]);
          if (!blu->insert(pnt))
            ptc->psh=-ANIPATCH_PUSH; // kill patch
        } else { // deformed
          ptc->psh=-ANIPATCH_PUSH; // kill patch
          // test if there is room for a new patch at this position:
          //pnt.as( (double)msh->arr.data[v3f+0],
          //        (double)msh->arr.data[v3f+1]);
          //if (blu->insert(pnt)) {
          //  aniPatch* tmp=new aniPatch(str,blu->rad,pnt,ord++,lev);
          //  list.add(tmp);
          //}
        }
      ptc->alp+=ptc->psh;
      ptc->alp=min(max(-1,ptc->alp),1);
      if (0<=ptc->alp)
        list.add(ptc);
      else
        delete ptc;
    }
    // fill gaps with new patches
    int num=blu->point.numel; // old number of points
    blu->complete(4);
    int chk;
    for (c0=num+1;c0<=blu->point.numel;++c0) {
      chk=blu->point.get(c0,pnt);
      assert(chk);
      list.add(new aniPatch(str,blu->rad,pnt,ord++,lev));
    }
    blu->empty();
    opengl_bulletin.add(tensor<double>::string( (time::now()-tic).ms() ));  
  }
}

void aniFlow::draw(void) {
  int c0;
  glNormal3d(0,0,1);
  tensor<double> ext=blu->extension();
  double h=-0.01;
  glBegin(GL_QUADS);
    glVertex3d( 0, 0, h);
    glVertex3d(ext(0), 0, h);
    glVertex3d(ext(0),ext(1), h);
    glVertex3d( 0, ext(1), h);
  glEnd();

  if (keyboard(' ')) {    
    tensor<double> pnt(1,4);
    tensor<double> dir(1,4);
    tensor<double> itp(1,2);    
    glBegin(GL_LINES);
      for (int c0=0;c0<fld[0].size[0];c0+=3)
      for (int c1=0;c1<fld[0].size[1];c1+=3) {
        itp.as( c0/(double)fld[0].size[0],
                c1/(double)fld[1].size[1]);
        pnt.as( itp.data[0]*ext(0),
                itp.data[1]*ext(1),
                0.0,1.0);
        dir.as( fld[0].interp(itp)*0.3,
                fld[1].interp(itp)*0.3,
                0.0, 0.0);
        glColor3d( 0.5, 0.5, 0.5);
        glVertex3dv( pnt.data     );
        glColor3d(1,1,1);
        glVertex3dv((pnt+dir).data);
      }
    glEnd();
  }
  
  tensor<aniPatch*> crt=list.toTensor(); // list to tensor
  tensor<int> nrm(1,crt.numel);     
  for (c0=0;c0<crt.numel;++c0)
    nrm(c0)=crt(c0)->ord;
  tensor<int> ind;
  nrm.sort(TENSOR_SORT_ABC,ind);
  
  glPushAttrib(GL_ENABLE_BIT);
	  glEnable(GL_BLEND);
	  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    for (c0=0;c0<crt.numel;++c0)
      crt(ind(c0))->draw(); // draw all patches in list ordered
    
    glBlendFunc(GL_ONE,GL_ONE);
  glPopAttrib();
  
}
