// code by jph

#ifndef AUDIO_INCLUDED
#define AUDIO_INCLUDED

#include "../_extern/fmod.h"

static char AUDIO_BUFFER[128];


class audio {
public:
  audio();  
  ~audio();
  audio(char* name,uint mode);
  void play();
  void doppler(float fac);
  char* toInfo();
  
  tensor<double> pos; // vector
  tensor<double> vel; // vector
  float vol;
  float private_frq; // original sampling rate of audio sample
  FMOD_CHANNEL* channel;
  FMOD_SOUND* sound; // maybe not needed!
// -----------------------------------------------------------------  
  static void main();
  static void setVector(FMOD_VECTOR* myVector,tensor<double> myTensor);
  static void listen(tensor<double>& pos,tensor<double>& vel);
  
  static FMOD_SYSTEM* system;
  static FMOD_SOUNDGROUP* master;
  static hash<tensor<char>,FMOD_SOUND*> hash;
};

audio::audio() {
  channel=0;
}

audio::~audio() {
  if (channel)
    FMOD_Channel_Stop(channel);
}

audio::audio(char* name,uint mode) { // FMOD_LOOP_NORMAL | FMOD_3D
  tensor<char> key=tensor<char>::string(name);
  if (!hash.get(key,sound)) {
    FMOD_RESULT res=FMOD_System_CreateSound(system,name,mode,0,&sound);      
    if (res!=FMOD_OK)
      printf("!!! ERROR %s\n",name);
    else {
      float tmp_vol=0;
      float pan=0;
      int   pty=0;
      FMOD_Sound_GetDefaults(sound,&private_frq,&tmp_vol,&pan,&pty);
      //printf("%s ->\t%s %10.2e %6.3f %6.3f %i\n",str,private_frq,tmp_vol,pan,pty);
      printf("%s ->\t%s\n",name,toInfo());
    }
  }
  FMOD_System_PlaySound(system,FMOD_CHANNEL_FREE,sound,1,&channel);
  pos=tensor<double>::vector(4);
  vel=tensor<double>::vector(4);
  vol=1.0;
  if (mode&FMOD_LOOP_NORMAL) {
    uint len;
    FMOD_Sound_GetLength(sound,&len,FMOD_TIMEUNIT_MS);
    FMOD_Channel_SetPosition(channel,random<int>::uniform(len),FMOD_TIMEUNIT_MS);
  }
}

void audio::play() {
  FMOD_VECTOR pnt,vlc;
  setVector(&pnt,pos);
  setVector(&vlc,vel);
  FMOD_Channel_Set3DAttributes(channel,&pnt,&vlc);
  FMOD_Channel_SetVolume(channel,vol);
  FMOD_Channel_SetPaused(channel,0);
}

void audio::doppler(float fac) {
  FMOD_Channel_SetFrequency(channel,private_frq*fac);    
}

char* audio::toInfo() {
  sprintf(AUDIO_BUFFER,"audio(%iHz,...)",(int)private_frq);
  return AUDIO_BUFFER;
}

void audio::main() {
  FMOD_System_Create(&system);
  //uint ver;
  //FMOD_System_GetVersion(system,&ver);
  //printf("fmod %i\n",ver);
  FMOD_System_Init(system,1024,FMOD_INIT_3D_RIGHTHANDED|FMOD_INIT_VOL0_BECOMES_VIRTUAL,0);    
  FMOD_System_GetMasterSoundGroup(system,&master);
}

void audio::setVector(FMOD_VECTOR* myVector,tensor<double> myTensor) {
  myVector->x=myTensor.data[0];
  myVector->y=myTensor.data[1];
  myVector->z=myTensor.data[2];
}

void audio::listen(tensor<double>& pos,tensor<double>& vel) {
  FMOD_VECTOR pnt,fwd,upw,vlc;
  setVector(&upw,pos[1]);
  setVector(&fwd,pos[2]);
  setVector(&pnt,pos[3]);
  setVector(&vlc,vel);
  FMOD_System_Set3DListenerAttributes(system,0,&pnt,&vlc,&fwd,&upw);     // update 'ears'
  FMOD_System_Update(system);
}


FMOD_SYSTEM* audio::system=0;
FMOD_SOUNDGROUP* audio::master=0;
hash<tensor<char>,FMOD_SOUND*> audio::hash(128);

#endif
