// code by jph
// adapted from joyride

#ifndef CONTROL_INCLUDED
#define CONTROL_INCLUDED

#define     CONTROL_EPS   1e-6

static char CONTROL_BUFFER[1024];

// TODO this class needs careful revision and testing
// with regards to automated control !!!

template<class type> // is template necessary !?! 
class control {
public:

  // d_t pn = vn
  // d_t vn = MU * ( CN - vn )

  static type dt;

  type    CN;         // IN max speed (= max accel)
  type    MU;         // IN strength of boost and friction
  type    EP;         // IN max error for no action
  type    EV;         // IN max error for no action
  type    vd;         // IN velocity desired
  type    pd;         // IN position desired (not used in velocity control)
  type    vn;         // IN/OUT velocity
  type    pn;         // IN/OUT position
  type    cn;         // IN/OUT control
  
  // control<double> myControl=new control<double>(1,1,.1,.1);
  
  control<type>(void) {
    CN=1;
    MU=1;
    EP=1e-2;
    EV=1e-2;
    vd=0;
    pd=0;
    vn=0;
    pn=0;
    cn=0;    
  }
  
  // BEGIN: Euler integration
  
  void integrate(void) { // ok
    vn+=MU*(cn-vn)*dt;
    pn+=vn*dt;
  }
  
  type V() { // ok
    return cn=math::Clip<type>(-CN, (vd-vn*(1-MU*dt))/(dt*MU) ,CN);
  }
  
  // V(type plo,type phi) works but avoid if possible: e.g. use vd=0 pd=plo/phi instead!
  void V(type plo,type phi) { // ok
    cn=math::Clip<type>(cVP(vn,0,pn,plo),V(),cVP(vn,0,pn,phi)); // tested with cat
  }
  
  void VP() { // ok, safe for vd=0; in general: CAUTION: simple integration might lead to bang-bang-bang-...
    cn=cVP(vn,vd,pn,pd);
  }
  
  type cVP(type vn,type vd,type pn,type pd) {
    if (CONTROL_EPS<abs(vd-vn)||CONTROL_EPS<abs(pd-pn)) {
      type ct=CN*math::Sign<type>(vd-vn);
      return math::Sign<type>(pd-(pn+(vn-vd+ct*log((ct-vn)/(ct-vd)))/MU))*min((abs(vd-vn)+abs(pd-pn))/(2*dt*MU),CN);
    } else
      return V();
  }
  
  int VPerror(void) {
    if (EP<abs(pd-pn)||EV<abs(vd-vn))
      return 1;
    else
      return 0;
  }
  
  // END: Euler integration
  
  // -----------------------------------------------------------------
  
  // BEGIN: exact integration
  
  void EXACT_integrate(void) { // ok
    type etm=exp(-dt*MU);
    pn=pn+vn*(1-etm)/MU-cn*(1-dt*MU-etm)/MU;
    vn=vn*etm+cn*(1-etm);
  }

  type EXACT_V() { // ok
    type etm=exp( dt*MU);
    return cn=math::Clip<type>(-CN,(etm*vd-vn)/(etm-1),CN);
  }
  
  // EXACT_V(type plo,type phi) works but avoid if possible: e.g. use vd=0 pd=plo/phi instead!
  type EXACT_V(type plo,type phi) { // ok
    return cn=math::Clip<type>(EXACT_cVP(vn,0,pn,plo),EXACT_V(),EXACT_cVP(vn,0,pn,phi)); // tested with cat
  }
  
  type EXACT_VP() { // ok
    return cn=EXACT_cVP(vn,vd,pn,pd);
  }
  
  type EXACT_cVP(type vn,type vd,type pn,type pd) { // ok
    if (CONTROL_EPS<abs(vd-vn)||CONTROL_EPS<abs(pd-pn)) {
      type ct=CN*math::Sign<type>(vd-vn);
      return math::Sign<type>(pd-(pn+(vn-vd+ct*log((ct-vn)/(ct-vd)))/MU))*min((abs(vd-vn)+abs(pd-pn))/(2*dt*MU),CN); // TODO 
    } else
      return EXACT_V();
  }

  // END: exact integration
  
  tensor<type> toTensor(void) {
    return tensor<type>::vector(5).as(cn,vn,pn,vd,pd);
  }

  char* toString(void) {    
    sprintf(CONTROL_BUFFER,"%12.4e,%12.4e,%12.4e -> %12.4e,%12.4e",cn,vn,pn,vd,pd);
    return CONTROL_BUFFER;
  }
  
};

float control<float>::dt=0.1;
double control<double>::dt=0.001;
  
#endif
