// code by jph

// since there is no copyconstructor, or operator= defined
//  => always use  const rapid&  when passing!

#include "../_extern/RAPID.H"

// static char RAPID_BUFFER[128];

// TODO only simply point(s) for connected components!!!

class rapid_touch {
public:
  int            num; // number of contacts
  //tensor<int>    par; // 2 x num with indices to triangles
  tensor<double> pnt;
  tensor<double> nrm;
  double         str; // 
};

class rapid {
public:
  rapid(void);
  ~rapid(void);
  void set(meshgl* msh);
  
  RAPID_model* model;
  tensor<double> are; // volume of triangle to compute weighted averages
  tensor<double> V3D; // 3 x T+2Q avg points of shapes
  tensor<double> N3D; // 3 x T+2Q avg normal or shapes

  static int collide(
    tensor<double>& pos0,const rapid& rap0,
    tensor<double>& pos1,const rapid& rap1,
    rapid_touch& tch);
};

rapid::rapid(void) { 
  model=0; // init here
}
  
inline rapid::~rapid(void) {
  if (model)
    delete model;
}

#define RAPID_SET                               \
    model->AddTri(v0.data,v1.data,v2.data,c1);  \
    crs=(v1-v0).x(v2-v0);                       \
    are(c1)=crs.norm()*0.5;                     \
    V3D.in(c1*3,((v0+v1+v2)*MATH_1_3).data,3);  \
    N3D.in(c1*3,         crs.normal().data,3);  \
    ++c1;


void rapid::set(meshgl* msh) {
  if (model)
    delete model;
  model=new RAPID_model();
  model->BeginModel();
  int num=msh->ind[MESH_TRIA].size[1] + msh->ind[MESH_QUAD].size[1]*2;
  are=tensor<double>(1,num);
  V3D=tensor<double>(2,3,num);
  N3D=tensor<double>(2,3,num);
  int ofs=msh->offsetV3F();
  int* dat;
  tensor<double> v0(1,3);
  tensor<double> v1(1,3);
  tensor<double> v2(1,3);
  tensor<double> crs;
  int c1=0;
  for (int c0=0;c0<msh->ind[MESH_TRIA].numel;c0+=3) {
    dat=&msh->ind[MESH_TRIA].data[c0];
    v0.in(0,&msh->arr[dat[0]].to<double>().data[ofs],3);
    v1.in(0,&msh->arr[dat[1]].to<double>().data[ofs],3);
    v2.in(0,&msh->arr[dat[2]].to<double>().data[ofs],3);
    RAPID_SET
  }
  for (int c0=0;c0<msh->ind[MESH_QUAD].numel;c0+=4) {
    dat=&msh->ind[MESH_QUAD].data[c0];    
    v0.in(0,&msh->arr[dat[0]].to<double>().data[ofs],3);
    v1.in(0,&msh->arr[dat[1]].to<double>().data[ofs],3);
    v2.in(0,&msh->arr[dat[2]].to<double>().data[ofs],3);
    RAPID_SET
    v0.in(0,&msh->arr[msh->ind[MESH_QUAD].data[c0+2]].to<double>().data[ofs],3);
    v1.in(0,&msh->arr[msh->ind[MESH_QUAD].data[c0+3]].to<double>().data[ofs],3);
    v2.in(0,&msh->arr[msh->ind[MESH_QUAD].data[c0+0]].to<double>().data[ofs],3);
    RAPID_SET
  }
  double res=are.min_(c1);
  model->EndModel();
}

int rapid::collide(tensor<double>& pos0,const rapid& rap0,
                   tensor<double>& pos1,const rapid& rap1,
                   rapid_touch &tch) {
  assert(rap0.model&&rap1.model);
  assert(rap0.model!=rap1.model); // TODO why is this a criteria
  int c0,id0,id1;
  double rot0[3][3];
  double rot1[3][3];
  rot0[0][0]=pos0.data[ 0]; rot0[0][1]=pos0.data[ 4]; rot0[0][2]=pos0.data[ 8];
  rot0[1][0]=pos0.data[ 1]; rot0[1][1]=pos0.data[ 5]; rot0[1][2]=pos0.data[ 9];
  rot0[2][0]=pos0.data[ 2]; rot0[2][1]=pos0.data[ 6]; rot0[2][2]=pos0.data[10];
  rot1[0][0]=pos1.data[ 0]; rot1[0][1]=pos1.data[ 4]; rot1[0][2]=pos1.data[ 8];
  rot1[1][0]=pos1.data[ 1]; rot1[1][1]=pos1.data[ 5]; rot1[1][2]=pos1.data[ 9];
  rot1[2][0]=pos1.data[ 2]; rot1[2][1]=pos1.data[ 6]; rot1[2][2]=pos1.data[10];
  RAPID_Collide(rot0,pos0[3].data,rap0.model,
                rot1,pos1[3].data,rap1.model,RAPID_ALL_CONTACTS);
  tch.num=RAPID_num_contacts;
  //tch.par=tensor<int>(2,2,RAPID_num_contacts);
  //int c1=-1;
  //for (c0=0;c0<RAPID_num_contacts;++c0) {
  //  tch.par.data[++c1]=RAPID_contact[c0].id1;
  //  tch.par.data[++c1]=RAPID_contact[c0].id2;
  //}
  tensor<double>::valid=1;
  tch.str=0;
  if (tch.num) {
    double sca;
    tensor<double> nrm0;
    tensor<double> nrm1;
    tensor<double> aff(1,4); aff(3)=1;
    tensor<double> vec0(1,4);
    tensor<double> vec1(1,4);
    tensor<double> pnt(1,4);
    tensor<double> nrm(1,4);
    for (c0=0;c0<RAPID_num_contacts;++c0) {
      id0=RAPID_contact[c0].id1;
      id1=RAPID_contact[c0].id2;
      if (rap0.are.data[id0]<rap1.are.data[id1]) { // smaller triangle contributes to position
        assert(id0*3<rap0.V3D.numel);
        pnt+=pos0*aff.in(0,&rap0.V3D.data[id0*3],3);
      } else {
        assert(id1*3<rap1.V3D.numel);
        pnt+=pos1*aff.in(0,&rap1.V3D.data[id1*3],3);
      }
      nrm0=pos0*vec0.in(0,&rap0.N3D.data[id0*3],3);
      nrm1=pos1*vec1.in(0,&rap1.N3D.data[id1*3],3);
      sca=-min(0,nrm0|nrm1)*min(rap0.are.data[id0],rap1.are.data[id1]);
      nrm+=(nrm1-nrm0)*sca;
      tch.str+=sca;
    }
    tch.pnt=pnt*(1/pnt(3));
    tch.nrm=nrm.normal();
  }
  return tch.num&&tensor<double>::valid;
}

